//
//  AudioOutput.m
//  AudioBoard
//  This code is released under the BSD License
///  Copyright (c) 2013 retronyms. All rights reserved.
//

#import "AudioIO.h"
#import <AudioToolbox/AudioToolbox.h>
#import <AVFoundation/AVAudioSession.h>

@implementation AudioIO
@synthesize delegate;
@synthesize isPlaying;
@synthesize isRecording;
@synthesize interAppConnected;
@synthesize currentSample;
#ifdef AUDIOBUS
@synthesize audiobusOutput;
@synthesize audiobusInput;
@synthesize audiobusAudioUnitWrapper;
@synthesize audiobusController;
#endif
@synthesize cpuPercent;
static float phase = 0;
OSStatus render(
                    void *inRefCon,
                    AudioUnitRenderActionFlags 	*ioActionFlags,
                    const AudioTimeStamp 		*inTimeStamp,
                    UInt32 						inBusNumber,
                    UInt32 						inNumberFrames,
                    AudioBufferList 			*ioData)

{
	AudioIO *audio = (__unsafe_unretained AudioIO *)inRefCon;

    audio->lastCallbackTime = CFAbsoluteTimeGetCurrent();
    float *input[2];
    float *output[2];
    output[0] = (float*)ioData->mBuffers[0].mData;
    output[1] = (float*)ioData->mBuffers[1].mData;
    if (audio->lastInputBuffer)
    {
        input[0] = (float*)audio->lastInputBuffer->mBuffers[0].mData;
        input[1] = (float*)audio->lastInputBuffer->mBuffers[1].mData;
        memcpy(input[1],input[0],inNumberFrames*sizeof(float)); //mono input on ios
    }
    else
    {
        memset(output[0], 0, sizeof(float)*inNumberFrames);
        memset(output[1], 0, sizeof(float)*inNumberFrames);
        input[0]=output[0];
        input[1]=output[1];
    }
    float increment = 440 * ((2*3.14)/44100.0);
    float sampleRate = [[AVAudioSession sharedInstance] sampleRate];
    for (int i=0;i<inNumberFrames;i++)
    {
        output[0][i] = sinf(phase);
        output[1][i] = output[0][i];
        phase += increment;
    }
    CFAbsoluteTime now = CFAbsoluteTimeGetCurrent();
    audio->cpuPercent = (now-audio->lastCallbackTime)/(inNumberFrames/sampleRate);
	return noErr;
}
static OSStatus recordingCallback(void *inRefCon,
                                  AudioUnitRenderActionFlags *ioActionFlags,
                                  const AudioTimeStamp *inTimeStamp,
                                  UInt32 inBusNumber,
                                  UInt32 inNumberFrames,
                                  AudioBufferList *ioData) {
    
    //make sure we have an allocated buffer of the right size.
    AudioIO *audio = (__bridge AudioIO *)inRefCon;
    if (!audio->lastInputBuffer || audio->lastInputBuffer->mBuffers[0].mDataByteSize != inNumberFrames * sizeof(Float32))
    {
        int kNumChannels = 2;
        audio->lastInputBuffer = (AudioBufferList*)malloc(sizeof(AudioBufferList) * kNumChannels);
        audio->lastInputBuffer->mNumberBuffers = kNumChannels; // 2 for stereo, 1 for mono
        for(int i = 0; i < 2; i++) {
            audio->lastInputBuffer->mBuffers[i].mNumberChannels = 1;
            audio->lastInputBuffer->mBuffers[i].mDataByteSize = inNumberFrames * sizeof(Float32);
            audio->lastInputBuffer->mBuffers[i].mData = (Float32*)malloc(sizeof(Float32) * inNumberFrames);
        }
    }
    // Obtain recorded samples
    OSStatus status;
    status = AudioUnitRender(audio->remoteIOUnit,
                             ioActionFlags,
                             inTimeStamp,
                             inBusNumber,
                             inNumberFrames,
                             audio->lastInputBuffer);
    return noErr;
}
+ (void) setAudioSessionActive {
    AVAudioSession *session = [AVAudioSession sharedInstance];
    NSError *err;
    [session setCategory: AVAudioSessionCategoryPlayAndRecord withOptions: AVAudioSessionCategoryOptionMixWithOthers error:  &err];
    [session setActive: YES error:  &err];
}

+ (void) setAudioSessionInActive {
    AVAudioSession *session = [AVAudioSession sharedInstance];
    NSError *err;
    [session setActive: NO error:  &err];
}

- (UIImage *)hostIcon
{
    if ([AudioIO interAppReady])
    {
        return AudioOutputUnitGetHostIcon(remoteIOUnit, 114);
    }
    return nil;
}
-(void) gotoHost
{
    if ([AudioIO interAppReady])
    {
        CFURLRef instrumentUrl;
        UInt32 dataSize = sizeof(instrumentUrl);
        OSStatus result = AudioUnitGetProperty(remoteIOUnit, kAudioUnitProperty_PeerURL, kAudioUnitScope_Global, 0, &instrumentUrl, &dataSize);
        if (result == noErr)
            [[UIApplication sharedApplication] openURL:(NSURL*)instrumentUrl];
    }
}

-(void) appHasGoneInBackground {
    inForeground = NO;
}
- (Float64)getCurrentSample
{
    [self updateTransportInfo];
    return outCurrentSampleInTimeLine;
}
- (void)updateTransportInfo
{
    if ([AudioIO interAppReady])
    {
        UInt32 dataSize = sizeof(HostCallbackInfo);
        HostCallbackInfo callBackInfo;
        OSStatus result = AudioUnitGetProperty(remoteIOUnit, kAudioUnitProperty_HostCallbacks, kAudioUnitScope_Global, 0, &callBackInfo, &dataSize);
        void * hostUserData = callBackInfo.hostUserData;
        Boolean playing;
        Boolean recording;
        
        result =  callBackInfo.transportStateProc2( hostUserData,
                                                            &playing,
                                                            &recording, NULL,
                                                            &outCurrentSampleInTimeLine,
                                                            NULL, NULL, NULL);
        isPlaying = playing;
        isRecording = recording;
    }
}
-(void) appHasGoneForeground {
    inForeground = YES;
    
    [AudioIO setAudioSessionActive];
    if (!interAppConnected)
    {
        OSStatus err = AudioOutputUnitStart(remoteIOUnit);
        NSLog(@"started audio after foregrounding, err: %d", (int)err);
    }
}
- (id)initWithAudioComponentDescription:(AudioComponentDescription)desc name:(NSString*)n
{
    self = [super init];
	if (self != nil)
    {
        name = n;
        [name retain];
        componentDescription = desc;
    }
    return self;
}
- (void)dealloc
{
    [super dealloc];
}

+ (BOOL)interAppReady
{
    NSArray *versionCompatibility = [[UIDevice currentDevice].systemVersion componentsSeparatedByString:@"."];
    if (versionCompatibility && versionCompatibility.count)
    {
        NSNumber *version = [versionCompatibility objectAtIndex:0];
        if (version && version.integerValue > 6)
        {
            return YES;
        }
    }
    return NO;
}

#define kInputBus 1
#define kOutputBus 0
- (void)startIO:(BOOL)publish
{
    NSLog(@"set session active");
    [AudioIO setAudioSessionActive];

    UIApplicationState appstate = [UIApplication sharedApplication].applicationState;
    inForeground = (appstate != UIApplicationStateBackground);
    
    [[NSNotificationCenter defaultCenter] addObserver: self
                                             selector: @selector(appHasGoneInBackground)
                                                 name: UIApplicationDidEnterBackgroundNotification
                                               object: nil];
    
    [[NSNotificationCenter defaultCenter] addObserver: self
                                             selector: @selector(appHasGoneForeground)
                                                 name: UIApplicationWillEnterForegroundNotification
                                               object: nil];

    

    // Configure the search parameters to find the default playback output unit
    // (called the kAudioUnitSubType_RemoteIO on iOS but
    // kAudioUnitSubType_DefaultOutput on Mac OS X)
    AudioComponentDescription defaultOutputDescription;
    defaultOutputDescription.componentType = kAudioUnitType_Output;
    defaultOutputDescription.componentSubType = kAudioUnitSubType_RemoteIO;
    defaultOutputDescription.componentManufacturer = kAudioUnitManufacturer_Apple;
    defaultOutputDescription.componentFlags = 0;
    defaultOutputDescription.componentFlagsMask = 0;
    
    // Get the default playback output unit
    AudioComponent defaultOutput = AudioComponentFindNext(NULL, &defaultOutputDescription);
    NSAssert(defaultOutput, @"Can't find default output");
    
    // Create a new unit based on this that we'll use for output
    AudioComponentInstanceNew(defaultOutput, &remoteIOUnit);
    // Enable IO for recording
    UInt32 flag = 1;
    OSStatus status = AudioUnitSetProperty(remoteIOUnit,
                                  kAudioOutputUnitProperty_EnableIO,
                                  kAudioUnitScope_Input,
                                  kInputBus,
                                  &flag,
                                  sizeof(flag));
    
    // Enable IO for playback
    status = AudioUnitSetProperty(remoteIOUnit,
                                  kAudioOutputUnitProperty_EnableIO,
                                  kAudioUnitScope_Output,
                                  kOutputBus,
                                  &flag,
                                  sizeof(flag));
    
    

    
    // Set the format to 32 bit, single channel, floating point, linear PCM
    const int four_bytes_per_float = 4;
    const int eight_bits_per_byte = 8;
    AudioStreamBasicDescription streamFormat;
    streamFormat.mSampleRate = [[AVAudioSession sharedInstance] sampleRate]; //hardware samplerate
    streamFormat.mFormatID = kAudioFormatLinearPCM;
    streamFormat.mFormatFlags =
    kAudioFormatFlagsNativeFloatPacked | kAudioFormatFlagIsNonInterleaved;
    streamFormat.mBytesPerPacket = four_bytes_per_float;
    streamFormat.mFramesPerPacket = 1;
    streamFormat.mBytesPerFrame = four_bytes_per_float;
    streamFormat.mChannelsPerFrame = 2;
    streamFormat.mBitsPerChannel = four_bytes_per_float * eight_bits_per_byte;

//                             | i                   o |
//    -- BUS 1 -- from mic --> | n    REMOTE I/O     u | -- BUS 1 -- to app -->
//                             | p      AUDIO        t |
//    -- BUS 0 -- from app --> | u       UNIT        p | -- BUS 0 -- to speaker -->
//                             | t                   u |
//                             |                     t |
//                             -------------------------

    // Tell Remote IO what formats we want to use in our app.
    status = AudioUnitSetProperty(remoteIOUnit,
                                  kAudioUnitProperty_StreamFormat,
                                  kAudioUnitScope_Input,
                                  kOutputBus,
                                  &streamFormat,
                                  sizeof(streamFormat));

    status = AudioUnitSetProperty(remoteIOUnit,
                                  kAudioUnitProperty_StreamFormat,
                                  kAudioUnitScope_Output,
                                  kInputBus,
                                  &streamFormat,
                                  sizeof(streamFormat));

    
    
    // Set input callback
    AURenderCallbackStruct callbackStruct;
    callbackStruct.inputProc = recordingCallback;
    callbackStruct.inputProcRefCon = self;
    status = AudioUnitSetProperty(remoteIOUnit,
                                  kAudioOutputUnitProperty_SetInputCallback,
                                  kAudioUnitScope_Global,
                                  kInputBus,
                                  &callbackStruct,
                                  sizeof(callbackStruct));
    
    // Set output callback
    callbackStruct.inputProc = render;
    callbackStruct.inputProcRefCon = self;
    status = AudioUnitSetProperty(remoteIOUnit,
                                  kAudioUnitProperty_SetRenderCallback,
                                  kAudioUnitScope_Global,
                                  kOutputBus,
                                  &callbackStruct, 
                                  sizeof(callbackStruct));
    
    
    lastCallbackTime = 0;
    if ([AudioIO interAppReady] && publish)
    {
        OSStatus result = AudioOutputUnitPublish(&componentDescription, (CFStringRef)name, 1,remoteIOUnit);
        if (result != noErr)
            NSLog(@"AudioOutputUnitPublish instrument result: %d", (int)result);
    }
    
    NSLog(@"initializing audio");
    status = AudioUnitInitialize(remoteIOUnit);
    NSLog(@"done initializing audio, status: %d", (int)status);

    if(inForeground)
    { // If we're in the background, we've been launched by interapp audio,
      // and we need to hold off on starting until the connected callback comes in.
        NSLog(@"done setting session active");
        NSLog(@"starting audio");
        status = AudioOutputUnitStart(remoteIOUnit);
        NSLog(@"done starting audio, status: %d", (int)status);
        if (status)
        {
            fprintf(stderr,"start error\n");
        }
    }
    [self addAudioUnitPropertyListener];
    [self updateTransportInfo];
    if ([AudioIO interAppReady])
    {
        UInt32 connect;
        UInt32 dataSize = sizeof(UInt32);
        AudioUnitGetProperty(remoteIOUnit, kAudioUnitProperty_IsInterAppConnected, kAudioUnitScope_Global, 0, &connect, &dataSize);
        interAppConnected = connect;
        NSLog(@"interAppConnected: %d", (int)connect);
    }
}
- (void)activateAudiobus:(NSString*)apikey url:(NSString*)launchurl
{
#ifdef AUDIOBUS
    audiobusController = [[ABAudiobusController alloc]
                          initWithAppLaunchURL:[NSURL URLWithString:launchurl]
                          apiKey:apikey];
    
    audiobusOutput = [audiobusController addOutputPortNamed:@"Audio Output"
                     title:NSLocalizedString(@"Main App Output", @"")];
    
    //create an input if you need it.
    
    audiobusAudioUnitWrapper = [[ABAudiobusAudioUnitWrapper alloc]
                                     initWithAudiobusController:audiobusController
                                     audioUnit:outputUnit
                                     output:audiobusOutput
                                     input:nil];
#endif
}

- (void)toggleInterAppPlay
{
    AudioUnitRemoteControlEvent controlEvent = kAudioUnitRemoteControlEvent_TogglePlayPause;
    UInt32 dataSize = sizeof(controlEvent);
    AudioUnitSetProperty(remoteIOUnit, kAudioOutputUnitProperty_RemoteControlToHost, kAudioUnitScope_Global, 0, &controlEvent, dataSize);
}
- (void)toggleInterAppRecord
{
    AudioUnitRemoteControlEvent controlEvent = kAudioUnitRemoteControlEvent_ToggleRecord;
    UInt32 dataSize = sizeof(controlEvent);
    AudioUnitSetProperty(remoteIOUnit, kAudioOutputUnitProperty_RemoteControlToHost, kAudioUnitScope_Global, 0, &controlEvent, dataSize);
    
}
void AudioUnitPropertyChangeDispatcher(void *inRefCon, AudioUnit inUnit, AudioUnitPropertyID inID, AudioUnitScope inScope, AudioUnitElement inElement)
{
    AudioIO *audio = (__bridge AudioIO *)inRefCon;
    NSObject *delobj = (NSObject *)audio.delegate;
    if (inID==kAudioUnitProperty_IsInterAppConnected)
    {
        UInt32 connect;
        UInt32 dataSize = sizeof(UInt32);
        AudioUnitGetProperty(inUnit, kAudioUnitProperty_IsInterAppConnected, kAudioUnitScope_Global, 0, &connect, &dataSize);
        if (connect)
        {
            audio->interAppConnected = connect;
            [audio updateTransportInfo];
            [AudioIO setAudioSessionActive];
            AudioOutputUnitStart(audio->remoteIOUnit);

            if (audio->delegate)
            {
                [delobj performSelectorOnMainThread:@selector(interAppConnected) withObject:nil waitUntilDone:NO];
            }
        }
        else
        {
            audio->interAppConnected = connect;
            [AudioIO setAudioSessionActive];
            if (audio.delegate)
            {
                [delobj performSelectorOnMainThread:@selector(interAppDisconnected) withObject:nil waitUntilDone:NO];
            }
        }
    }
    if (inID==kAudioOutputUnitProperty_HostTransportState)
    {
        [audio updateTransportInfo];
        if (audio.delegate)
        {
            [delobj performSelectorOnMainThread:@selector(interAppTransportChanged) withObject:nil waitUntilDone:NO];
        }
    }
}
-(void) setupMidiCallBacks:(AudioUnit*)output userData:(void*)inUserData
{
    AudioOutputUnitMIDICallbacks callBackStruct;
    callBackStruct.userData = inUserData;
    callBackStruct.MIDIEventProc = MIDIEventProcCallBack;
    callBackStruct.MIDISysExProc = NULL;
    AudioUnitSetProperty (*output,
                                kAudioOutputUnitProperty_MIDICallbacks,
                                kAudioUnitScope_Global,
                                0,
                                &callBackStruct,
                                sizeof(callBackStruct));
}

void MIDIEventProcCallBack(void *userData, UInt32 inStatus, UInt32 inData1, UInt32 inData2, UInt32 inOffsetSampleFrame)
{
    fprintf(stderr,"midi received from host\n");
}

-(void)addAudioUnitPropertyListener
{
    if ([AudioIO interAppReady])
    {
        AudioUnitRemovePropertyListenerWithUserData(remoteIOUnit, kAudioUnitProperty_IsInterAppConnected, AudioUnitPropertyChangeDispatcher, (__bridge void *)(self));
        AudioUnitRemovePropertyListenerWithUserData(remoteIOUnit, kAudioOutputUnitProperty_HostTransportState, AudioUnitPropertyChangeDispatcher, (__bridge void *)(self));

        AudioUnitAddPropertyListener(remoteIOUnit,
                                     kAudioUnitProperty_IsInterAppConnected,
                                     AudioUnitPropertyChangeDispatcher,
                                     (__bridge void *)(self));
        AudioUnitAddPropertyListener(remoteIOUnit,
                                     kAudioOutputUnitProperty_HostTransportState,
                                     AudioUnitPropertyChangeDispatcher,
                                     (__bridge void *)(self));
        [self setupMidiCallBacks:&remoteIOUnit userData:(__bridge void *)(self)];
    }
}
@end
